
include std/iup.e
include std/iup_scintilla.e
include std/cmdline.e
include std/console.e
include std/filesys.e
include std/io.e
include std/pipeio.e
include std/sequence.e
include std/wildcard.e
include euphoria/keywords.e

--**
-- N.B. THIS IS A BUG WORK-AROUND!
--
-- Keypress events (e.g. Ctrl+N) are somtimes triggered twice.
-- We use a timer to prevent our hotkey-enabled callbacks from
-- being called too many times within a small timeframe (50ms).
--
integer skip_next_event = 0

--**
-- Add a new tab.
--
public function add_tab( sequence title = "", sequence text = "" )

    atom sci = IupScintilla()
    IupSetAttributes( sci, "TABSIZE=4, EXPAND=YES, VISIBLE=YES" )
    IupSetAttribute( sci, "TABTITLE", title )

    atom tabs = IupGetHandle( "tabs" )
    IupAppend( tabs, sci )
	IupMap( sci )
    IupSetAttributeHandle( tabs, "VALUE", sci )
    IupSetFocus( sci )
	IupRefresh( tabs )

    IupSetAttribute( sci, "INSERT0", text )
	IupSetAttribute( sci, "UNDO", NULL )
 	IupSetAttribute( sci, "SAVEDSTATE", NULL )
    IupSetCallback( sci, "SAVEPOINT_CB", Icallback("savepoint_cb") )

    return sci
end function

--**
-- Apply Scintilla lexer configuration based on extension.
--
public procedure apply_lexer( atom sci, sequence path = "" )

	sequence name = filesys:filename( path )
	sequence list = { "*.e", "*.ew", "*.ex", "*.exw" }

	ifdef WINDOWS then
		IupSetAttribute( sci, "STYLEFONT32", "Consolas" )
	elsedef
		IupSetAttribute( sci, "STYLEFONT32", "Monospace" )
	end ifdef

	IupSetAttribute( sci, "STYLEFONTSIZE32", "10" )
	IupSetAttribute( sci, "STYLECLEARALL", "" )

	if length( path ) = 0 or wildcard_any( list, name ) then

		sequence keywords0 = stdseq:join( keywords:keywords, " " )
		sequence keywords1 = stdseq:join( keywords:builtins, " " )

		-- FIXME: need to implement real Euphoria lexer!
		IupSetAttribute( sci, "LEXERLANGUAGE", "lua" )

		-- apply Euphoria keyword style
		IupSetAttribute( sci, "KEYWORDS0", keywords0 )
		IupSetAttribute( sci, "STYLEFGCOLOR5", "0 0 255" ) -- blue

		-- apply Euphoria builtin style
		IupSetAttribute( sci, "KEYWORDS1", keywords1 )
		IupSetAttribute( sci, "STYLEFGCOLOR13", "255 0 255" ) -- magenta

		-- apply Euphoria comment style
		IupSetAttribute( sci, "STYLEFGCOLOR2", "255 0 0" ) -- red
		IupSetAttribute( sci, "STYLEBOLD2", "YES" )

		-- apply Euphoria string style
		IupSetAttribute( sci, "STYLEFGCOLOR6", "0 128 0" ) -- green
		IupSetAttribute( sci, "STYLEFGCOLOR7", "0 128 0" ) -- green

	end if

end procedure

--**
-- Match any of a list of wildcards.
--
public function wildcard_any( sequence list, sequence text )

	for i = 1 to length( list ) do

		if wildcard:is_match( list[i], text ) then
			return 1
		end if

	end for

	return 0
end function

--**
-- Create a new empty document.
--
public function new_cb( atom self )
--	IupAlarm( "Action", "New", "OK" )

	if skip_next_event then
		return IUP_IGNORE
	end if

	atom sci = add_tab( "Untitled" )
	IupSetAttribute( sci, "PATH", "" )
	apply_lexer( sci )

	atom status = IupGetHandle( "status" )
	IupSetAttribute( status, "TITLE", "" )

	skip_next_event = 1

	return IUP_DEFAULT
end function

--**
-- Open an existing document.
--
public function open_cb( atom self )
--	IupAlarm( "Action", "Open", "OK" )

	if skip_next_event then
		return IUP_IGNORE
	end if

	atom dlg = IupFileDlg()
	IupSetAttributes( dlg, "DIALOGTYPE=OPEN, TITLE=Open," &
		" FILTER=\"*.*\", FILTERINFO=\"All files (*.*)\"" )

	IupPopup( dlg, IUP_CURRENT, IUP_CURRENT )
	if IupGetInt( dlg, "STATUS" ) = IUP_INVALID then
		return IUP_IGNORE
	end if

	sequence path = IupGetAttribute( dlg, "VALUE" )
	sequence name = filesys:filename( path )
	sequence text = read_file( path )

	atom sci = add_tab( name, text )
	IupSetAttribute( sci, "PATH", path )
	apply_lexer( sci, path )

	atom status = IupGetHandle( "status" )
	IupSetAttribute( status, "TITLE", path )

	skip_next_event = 1

	return IUP_DEFAULT
end function

--**
-- Save the currently selected document.
--
public function save_cb( atom self )
--	IupAlarm( "Action", "Save", "OK" )

	if skip_next_event then
		return IUP_IGNORE
	end if

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetAttributeHandle( tabs, "VALUE" )

	sequence path = IupGetAttribute( sci, "PATH" )
	if length( path ) = 0 then
		-- file has not been named yet

		atom dlg = IupFileDlg()
		IupSetAttributes( dlg, "DIALOGTYPE=SAVE, TITLE=Save," &
			" FILTER=\"*.*\", FILTERINFO=\"All files (*.*)\"" )

		IupPopup( dlg, IUP_CURRENT, IUP_CURRENT )
		if IupGetInt( dlg, "STATUS" ) = IUP_INVALID then
			-- user cancelled
			return IUP_IGNORE
		end if

		path = IupGetAttribute( dlg, "VALUE" )
		IupSetAttribute( sci, "PATH", path )

	end if

	sequence text = IupGetAttribute( sci, "VALUE" )
	io:write_file( path, text )

	sequence name = filesys:filename( path )
	IupSetAttributeId( tabs, "TABTITLE", pos, name )
	IupSetAttribute( sci, "SAVEDSTATE" )

	skip_next_event = 1

	return IUP_DEFAULT
end function

--**
-- Save the document with a new name.
--
public function saveas_cb( atom self )
--	IupAlarm( "Action", "Save As", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom dlg = IupFileDlg()
	IupSetAttributes( dlg, "DIALOGTYPE=SAVE, TITLE=Save As," &
		" FILTER=\"*.*\", FILTERINFO=\"All files (*.*)\"" )

	IupPopup( dlg, IUP_CURRENT, IUP_CURRENT )
	if IupGetInt( dlg, "STATUS" ) = IUP_INVALID then
		-- user cancelled
		return IUP_DEFAULT
	end if

	atom sci = IupGetAttributeHandle( tabs, "VALUE" )

	sequence path = IupGetAttribute( dlg, "VALUE" )
	IupSetAttribute( sci, "PATH", path )

	sequence text = IupGetAttribute( sci, "VALUE" )
	io:write_file( path, text )

	sequence name = filesys:filename( path )
	IupSetAttributeId( tabs, "TABTITLE", pos, name )
	IupSetAttribute( sci, "SAVEDSTATE" )
	IupRefresh( tabs )

	return IUP_DEFAULT
end function

--**
-- Close the currently selected document.
--
public function close_cb( atom self )
--	IupAlarm( "Action", "Close", "OK" )

	if skip_next_event then
		return IUP_IGNORE
	end if

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetAttributeHandle( tabs, "VALUE" )
	atom modified = IupGetInt( sci, "SAVEDSTATE" )

	if modified then
		-- ask the user to save

		sequence name = "Untitled"
		sequence path = IupGetAttribute( sci, "PATH" )

		if length( path ) then
			name = filesys:filename( path )
		end if

		sequence message = sprintf( "Save changes to '%s'?", {name} )

		atom dlg = IupMessageDlg()
		IupSetAttribute( dlg, "TITLE", "Close" )
		IupSetAttribute( dlg, "VALUE", message )
		IupSetAttribute( dlg, "BUTTONS", "YESNOCANCEL" )
		IupSetAttribute( dlg, "DIALOGTYPE", "QUESTION" )

		IupPopup( dlg, IUP_CURRENT, IUP_CURRENT )
		atom response = IupGetInt( dlg, "BUTTONRESPONSE" )

		if response = 3 then
			-- user cancelled
			return IUP_IGNORE

		elsif response = 1 then
			-- user wants to save

			atom result = save_cb( self )
			if result != IUP_DEFAULT then
				return IUP_IGNORE
			end if

		end if

	end if

	-- remove the tab
	IupDestroy( sci )

	atom count = IupGetChildCount( tabs )
	if count != 0 then
		-- update the tab selection

		while pos >= count do
			-- pos is out of bounds
			pos -= 1
		end while

		sci = IupGetChild( tabs, pos )
		IupSetAttributeHandle( tabs, "VALUE", sci )

		IupRefresh( tabs )
		IupSetFocus( sci )

	end if

	skip_next_event = 1

	return IUP_DEFAULT
end function

--**
-- Exit the application.
--
public function exit_cb( atom self )
--	IupAlarm( "Action", "Exit", "OK" )

	-- tell the application to close
	return IUP_CLOSE
end function

--**
-- Undo the last bufferred action.
--
public function undo_cb( atom self )
--	IupAlarm( "Action", "Undo", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetAttributeHandle( tabs, "VALUE" )
	if IupGetInt( sci, "UNDO" ) then
		IupSetAttribute( sci, "UNDO", "YES" )
	end if

	return IUP_DEFAULT
end function

--**
-- Redo the last buffered action.
--
public function redo_cb( atom self )
--	IupAlarm( "Action", "Redo", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetAttributeHandle( tabs, "VALUE" )
	if IupGetInt( sci, "REDO" ) then
		IupSetAttribute( sci, "REDO", "YES" )
	end if

	return IUP_DEFAULT
end function

--**
-- Cut the currently selected text.
--
public function cut_cb( atom self )
--	IupAlarm( "Action", "Cut", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetChild( tabs, pos )
	IupSetAttribute( sci, "CLIPBOARD", "CUT" )

	return IUP_DEFAULT
end function

--**
-- Copy the currently selected text.
--
public function copy_cb( atom self )
--	IupAlarm( "Action", "Copy", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetChild( tabs, pos )
	IupSetAttribute( sci, "CLIPBOARD", "COPY" )

	return IUP_DEFAULT
end function

--**
-- Paste the contents of the clipboard.
--
public function paste_cb( atom self )
--	IupAlarm( "Action", "Paste", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetChild( tabs, pos )
	IupSetAttribute( sci, "CLIPBOARD", "PASTE" )

	return IUP_DEFAULT
end function

--**
-- Delete the currently selected text.
--
public function delete_cb( atom self )
--	IupAlarm( "Action", "Delete", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetChild( tabs, pos )
	IupSetAttribute( sci, "CLIPBOARD", "CLEAR" )

	return IUP_DEFAULT
end function

--**
-- Select all text in the current document.
--
public function selectall_cb( atom self )
--	IupAlarm( "Action", "Select All", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetChild( tabs, pos )
	atom count = IupGetInt( sci, "COUNT" )
	sequence selection = sprintf("%d:%d", {0,count} )
	IupSetAttribute( sci, "SELECTIONPOS", selection )

	return IUP_DEFAULT
end function

--**
-- Run the file in the current tab.
--
public function run_cb( atom self )
--	IupAlarm( "Action", "Run", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetInt( tabs, "VALUEPOS" )
	if pos = IUP_INVALID then
		-- nothing selected
		return IUP_IGNORE
	end if

	atom sci = IupGetChild( tabs, pos )
	sequence path = IupGetAttribute( sci, "PATH" )
	sequence name = filesys:filename( path )

	ifdef WINDOWS then
		sequence eui = "euiw.exe"
	elsedef
		sequence eui = "eui"
	end ifdef

	sequence cmd = build_commandline({ eui, path })

	chdir( filesys:pathname(path) )
	pipeio:exec( cmd, pipeio:create() )

	return IUP_DEFAULT
end function

--**
-- Show the "About" dialog.
--
public function about_cb( atom self )
--	IupAlarm( "Action", "About", "OK" )

	atom dialog = IupGetHandle( "dialog" )
	sequence title   = IupGetAttribute( dialog, "TITLE" )
	sequence version = IupGetAttribute( dialog, "VERSION" )
	sequence release = IupGetAttribute( dialog, "RELEASE" )
	sequence author  = IupGetAttribute( dialog, "AUTHOR" )
	sequence email   = IupGetAttribute( dialog, "EMAIL" )
	sequence website = IupGetAttribute( dialog, "WEBSITE" )

	sequence message = sprintf( "%s\n\t%s\n\t%s\n\nby %s\n\t%s\n\n%s",
		{title,version,release,author,email,website} )

	atom result = IupAlarm( title, message, "Close", "Website" )
	if result = 2 then
		-- user clicked website button
		IupHelp( website )
	end if

	return IUP_DEFAULT
end function

--**
-- Ask the user to save while closing.
--
public function closing_cb( atom self )
--	IupAlarm( "Action", "Closing", "OK" )

	atom tabs = IupGetHandle( "tabs" )

	atom count = IupGetChildCount( tabs )
	for pos = 0 to count - 1 do

		atom sci = IupGetChild( tabs, pos )
		atom modified = IupGetInt( sci, "SAVEDSTATE" )

		if modified then
			-- ask the user to save

			sequence name = "Untitled"
			sequence path = IupGetAttribute( sci, "PATH" )

			if length( path ) then
				name = filesys:filename( path )
			end if

			sequence message = sprintf( "Save changes to '%s'?", {name} )

			atom dlg = IupMessageDlg()
			IupSetAttribute( dlg, "TITLE", "Close" )
			IupSetAttribute( dlg, "VALUE", message )
			IupSetAttribute( dlg, "BUTTONS", "YESNOCANCEL" )
			IupSetAttribute( dlg, "DIALOGTYPE", "QUESTION" )

			IupPopup( dlg, IUP_CURRENT, IUP_CURRENT )
			atom response = IupGetInt( dlg, "BUTTONRESPONSE" )

			if response = 3 then
				-- user cancelled
				return IUP_IGNORE

			elsif response = 1 then
				-- user wants to save

				atom result = save_cb( self )
				if result != IUP_DEFAULT then
					return IUP_IGNORE
				end if

			end if

		end if

	end for

	return IUP_DEFAULT
end function

--**
-- Show layout dialog (Debugging).
--
public function layout_cb( atom self )
--	IupAlarm( "Action", "Layout", "OK" )

	atom dialog = IupGetHandle( "dialog" )
	atom layout = IupLayoutDialog( dialog )
	IupShowXY( layout, IUP_CURRENT, IUP_CURRENT )

	return IUP_DEFAULT
end function

--**
-- Reattach the toolbar to the dialog.
--
public function reattach_cb( atom self )
--	IupAlarm( "Action", "Reattach", "OK" )

	atom toolbar = IupGetHandle( "toolbar" )
	IupSetAttribute( toolbar, "RESTORE", NULL )

	return IUP_DEFAULT
end function

--**
-- Track the closure of the toolbar window.
--
public function detached_cb( atom self, atom dialog, atom x, atom y )
--	IupAlarm( "Action", "Detached", "OK" )

	IupSetCallback( dialog, "CLOSE_CB", Icallback("reattach_cb") )

	return IUP_DEFAULT
end function

--**
-- Set focus to the control when the tab changes.
--
public function tabchange_cb( atom self, atom new_tab, atom old_tab )

	IupSetFocus( new_tab )
	sequence path = IupGetAttribute( new_tab, "PATH" )

	atom status = IupGetHandle( "status" )
	IupSetAttribute( status, "TITLE", path )

	return IUP_DEFAULT
end function

--**
-- Remove the control when the tab is closed.
--
public function tabclose_cb( atom self, atom pos )
--	IupAlarm( "Action", "Tab close", "OK" )

	atom tabs = IupGetHandle( "tabs" )
	IupSetInt( tabs, "VALUE", pos )

	atom sci = IupGetAttributeHandle( tabs, "VALUE" )
	atom modified = IupGetInt( sci, "SAVEDSTATE" )

	if modified then
		-- ask the user to save

		sequence name = "Untitled"
		sequence path = IupGetAttribute( sci, "PATH" )

		if length( path ) then
			name = filesys:filename( path )
		end if

		sequence message = sprintf( "Save changes to '%s'?", {name} )

		atom dlg = IupMessageDlg()
		IupSetAttribute( dlg, "TITLE", "Close" )
		IupSetAttribute( dlg, "VALUE", message )
		IupSetAttribute( dlg, "BUTTONS", "YESNOCANCEL" )
		IupSetAttribute( dlg, "DIALOGTYPE", "QUESTION" )

		IupPopup( dlg, IUP_CURRENT, IUP_CURRENT )
		atom response = IupGetInt( dlg, "BUTTONRESPONSE" )

		if response = 3 then
			-- user cancelled
			return IUP_IGNORE

		elsif response = 1 then
			-- user wants to save

			atom result = save_cb( self )
			if result != IUP_DEFAULT then
				return IUP_IGNORE
			end if

		end if

	end if

	-- let the tab close
	return IUP_CONTINUE
end function

--**
-- Update the saved status on the tab.
--
public function savepoint_cb( atom self, atom status )
--	IupAlarm( "Action", "Save point", "OK" )

	sequence name = "Untitled"
	sequence path = IupGetAttribute( self, "PATH" )

	if length( path ) then
		name = filesys:filename( path )
	end if

	if not status then name &= "*" end if

	atom tabs = IupGetHandle( "tabs" )
	atom pos = IupGetChildPos( tabs, self )
	IupSetAttributeId( tabs, "TABTITLE", pos, name )
	IupRefresh( tabs )

	return IUP_DEFAULT
end function

--**
-- Reset the skip_next_event flag.
--
-- Update the menu and toolbar status.
--
public function timer_cb( atom self )
--	IupAlarm( "Action", "Timer", "OK" )

	skip_next_event = 0

	atom tabs = IupGetHandle( "tabs" )
	atom sci = IupGetAttributeHandle( tabs, "VALUE" )

	-- File menu
	toggle_item( "file_save", IupGetAttribute(sci, "SAVEDSTATE") )
	toggle_item( "file_saveas", (sci != NULL) )
	toggle_item( "file_close",  (sci != NULL) )

	-- Edit menu
	toggle_item( "edit_undo",      IupGetAttribute(sci, "UNDO") )
	toggle_item( "edit_redo",      IupGetAttribute(sci, "REDO") )
	-- --------
	toggle_item( "edit_cut",       IupGetAttribute(sci, "SELECTEDTEXT") )
	toggle_item( "edit_copy",      IupGetAttribute(sci, "SELECTEDTEXT") )
	toggle_item( "edit_paste",     IupGetAttribute(sci, "CLIPBOARD")    )
	toggle_item( "edit_delete",    IupGetAttribute(sci, "SELECTEDTEXT") )
	-- --------
	toggle_item( "edit_selectall", IupGetAttribute(sci, "COUNT") )

	-- Run menu
	toggle_item( "run_run", IupGetAttribute(sci, "PATH") )

	-- Toolbar
	toggle_item( "tool_save",   IupGetAttribute(sci, "SAVEDSTATE") )
	-- --------
	toggle_item( "tool_cut",    IupGetAttribute(sci, "SELECTEDTEXT") )
	toggle_item( "tool_copy",   IupGetAttribute(sci, "SELECTEDTEXT") )
	toggle_item( "tool_paste",  IupGetAttribute(sci, "CLIPBOARD")    )
	toggle_item( "tool_undo",   IupGetAttribute(sci, "UNDO")         )
	toggle_item( "tool_redo",   IupGetAttribute(sci, "REDO")         )

	return IUP_DEFAULT
end function

--**
-- Toggle item state based on an attribute value.
--
public procedure toggle_item( sequence handle, object value )

	if sequence( value ) then
		value = length(value) > 0
			and equal(value, "NO") = 0
			and equal(value,  "0") = 0
	end if

	atom item = IupGetHandle( handle )
	IupSetInt( item, "ACTIVE", value )

end procedure

--**
-- Register hotkey commands.
--
public procedure register_hotkeys()

	atom dialog = IupGetHandle( "dialog" )

	-- File menu shortcuts
	IupSetCallback( dialog, "K_cN",  Icallback("new_cb") )			-- Ctrl-N
	IupSetCallback( dialog, "K_cO",  Icallback("open_cb") )			-- Ctrl-O
	IupSetCallback( dialog, "K_cS",  Icallback("save_cb") )			-- Ctrl-S
	IupSetCallback( dialog, "K_cW",  Icallback("close_cb") )		-- Ctrl-W

	-- Edit menu shortcuts (handled by Scintilla control)
--	IupSetCallback( dialog, "K_cZ",  Icallback("undo_cb") )			-- Ctrl-Z
--	IupSetCallback( dialog, "K_cY",  Icallback("redo_cb") )			-- Ctrl-Y
--	IupSetCallback( dialog, "K_cX",  Icallback("cut_cb") )			-- Ctrl-X
--	IupSetCallback( dialog, "K_cC",  Icallback("copy_cb") )			-- Ctrl-C
--	IupSetCallback( dialog, "K_cV",  Icallback("paste_cb") )		-- Ctrl+V
--	IupSetCallback( dialog, "K_cA",  Icallback("selectall_cb") )	-- Ctrl+A

	-- Run menu shortcuts
	IupSetCallback( dialog, "K_F5",  Icallback("run_cb") )			-- F5

	-- Help menu shortcuts
	IupSetCallback( dialog, "K_F12", Icallback("about_cb") )		-- F12

end procedure

